import akka.actor._
import scala.math.BigInt.int2bigInt
import scala.collection.mutable.Buffer

object Bank {
  case object CreateAccount
  case class NewBankAccount(account: ActorRef)
}

class Bank extends Actor {
  import Bank._

  val accounts  = Buffer.empty[ActorRef]  

  def receive = {
  	case CreateAccount =>  
  	  val account = context.actorOf(Props[BankAccount], "account-" + accounts.length)  	  
  	  accounts += account
  	  sender ! NewBankAccount(account)
  	case transfer: WireTransfer.Transfer =>
  	  val wireTransfer = context.actorOf(Props[WireTransfer])
  	  wireTransfer.forward(transfer)
  }
}

object BankAccount {
  case class Deposit(amount: BigInt) {
    require(amount > 0)
  }
  case class Withdraw(amount: BigInt) {
    require(amount > 0)
  }  
  case object Done
  case class Failed(reason: String)
  
  case object RequestBalance
  case class Balance(amount: BigInt)
}

class BankAccount extends Actor {
  import BankAccount._
  var balance: BigInt = 0
  def receive = {
    case Deposit(amount) =>
      balance += amount
      sender ! Done
    case Withdraw(amount) if amount <= balance => 
      balance -= amount
      sender ! Done
    case Withdraw(amount) =>
      sender ! Failed("No sufficient balance")
    case RequestBalance =>
      sender ! Balance(balance)
    case _ => 
      sender ! Failed("WTF?")
  }
}

object WireTransfer {
  case class Transfer(from: ActorRef, to: ActorRef, amount: BigInt)
  case object Done
  case class Failed(reason: String)
}

class WireTransfer extends Actor {
  import WireTransfer._
  
  def receive = {
    case Transfer(from, to, amount) =>
      from ! BankAccount.Withdraw(amount)
      context.become(awaitWithdraw(to,amount,sender))    
  }
  
  def awaitWithdraw(to: ActorRef, amount: BigInt, client: ActorRef): Receive = {
    case BankAccount.Done =>
      to ! BankAccount.Deposit(amount)
      context.become(awaitDeposit(client))
    case BankAccount.Failed(reason) =>
      client ! Failed("withdraw failed: " + reason)
      context.stop(self)
  }
  
  def awaitDeposit(client: ActorRef): Receive = {
    case BankAccount.Done =>
      client ! Done
      context.stop(self)
  }
}

object BankTest extends App {
  import akka.pattern.{ ask, pipe }
  import akka.util.Timeout
  import scala.concurrent.duration._
  import scala.concurrent.ExecutionContext.Implicits.global

  val system = ActorSystem("bank-test")
  val bank = system.actorOf(Props[Bank],"bank")

  implicit val timeout = Timeout(5 seconds)

  for {
  	Bank.NewBankAccount(account1) <- bank ? Bank.CreateAccount
  	Bank.NewBankAccount(account2) <- bank ? Bank.CreateAccount
  	receipt <- account1 ? BankAccount.Deposit(10000)
  	receipt <- bank ? WireTransfer.Transfer(account1,account2,20000)
  	BankAccount.Balance(balance1) <- account1 ? BankAccount.RequestBalance
  	BankAccount.Balance(balance2) <- account2 ? BankAccount.RequestBalance
  } {
    println(receipt)
  	println(s"balances after transfer: account1: $balance1 account2: $balance2")
  	system.terminate()
  }
}