import akka.actor._
import scala.util.Random
import scala.concurrent.duration._
import scala.language.postfixOps

case object Take
case object Put
case object Ok
case object Refuse

object Philosopher {
  def props(leftChopstick: ActorRef, rightChopstick: ActorRef) = 
    Props(classOf[Philosopher], leftChopstick, rightChopstick)
}

object EvilPhilosopher {
	def props(leftChopstick: ActorRef, rightChopstick: ActorRef) =
		Props(classOf[EvilPhilosopher], leftChopstick, rightChopstick)
}

class EvilPhilosopher(leftChopstick: ActorRef, rightChopstick: ActorRef) extends Philosopher(leftChopstick,rightChopstick) {
	override def waitForLeftChopstick: Receive = ({
		case Refuse =>
			leftChopstick ! Put
			leftChopstick ! Take
	} : Receive) orElse super.waitForLeftChopstick

	override def waitForRightChopstick: Receive = ({
		case Refuse =>
			rightChopstick ! Put
			rightChopstick ! Take
	} : Receive) orElse super.waitForRightChopstick
}


class Philosopher(leftChopstick: ActorRef, rightChopstick: ActorRef) extends Actor with ActorLogging {
	implicit val executionContext = context.dispatcher

  def receive = waitForLeftChopstick

	case object FinishedThinking
	case object FinishedEating

	def eat() = {
  	log.info("eating")
  	context.system.scheduler.scheduleOnce(
		  200 millis, self, FinishedEating
	  )
		context.become(eating)
  }

  def think() = {
	  log.info("thinking")
	  context.system.scheduler.scheduleOnce(
		  100 millis, self, FinishedThinking
	  )
	  context.become(thinking)
  }

	def thinking: Receive = {
		case FinishedThinking =>
			leftChopstick ! Take
			context.become(waitForLeftChopstick)
	}

	def eating: Receive = {
		case FinishedEating =>
			leftChopstick ! Put
			rightChopstick ! Put
			think()
	}

  def waitForLeftChopstick: Receive = {
  	case Ok =>
  	  rightChopstick ! Take
		  context.become(waitForRightChopstick)
  	case Refuse =>
  	  think()
  }

  def waitForRightChopstick: Receive = {
  	case Ok =>
  	  eat()
  	case Refuse =>
  	  leftChopstick ! Put
  	  think()
  }

  override def preStart() = {
  	leftChopstick ! Take
  }
}

class Chopstick extends Actor with ActorLogging {
  def receive = available

  def available: Receive = {
  	case Take => 
  	  sender ! Ok
  	  context.become(taken(sender()))
  }

  def taken(owner: ActorRef): Receive = {
  	case Take =>
  	  sender ! Refuse
  	case Put if owner == sender =>
  	  context.become(available)
	  case Put =>
		  log.warning("there seems to be an evil philosopher.")
  }  
}

object Philosophers extends App {
  val n = 2
  val system = ActorSystem("philosophers")
  val chopsticks = List.tabulate(n)(i => 
  	system.actorOf(Props[Chopstick], "chopstick-" + i))
  val philosophers = List.tabulate(n) { i =>
	  val props = if (i == 0) EvilPhilosopher.props(chopsticks(i),chopsticks((i+1) % n))
	              else Philosopher.props(chopsticks(i),chopsticks((i+1) % n))
  	system.actorOf(props, "philosopher-" + i)
  }
  readLine()
  system.terminate()
}